# Copyright (c) 2016 by Kaminario Technologies, Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Volume driver for Kaminario K2 all-flash arrays."""

import math
import re
import threading
import time

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import units
from oslo_utils import versionutils
import requests
import six

import cinder
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import fields
from cinder import utils
from cinder.volume import configuration
from cinder.volume.drivers.san import san
from cinder.volume import volume_utils

krest = importutils.try_import("krest")

K2_MIN_VERSION = '2.2.0'
K2_LOCK_PREFIX = 'Kaminario'
MAX_K2_RETRY = 5
K2_REP_FAILED_OVER = fields.ReplicationStatus.FAILED_OVER
LOG = logging.getLogger(__name__)

kaminario_opts = [
    cfg.BoolOpt('auto_calc_max_oversubscription_ratio',
                default=False,
                help="K2 driver will calculate max_oversubscription_ratio "
                     "on setting this option as True."),
    cfg.BoolOpt('disable_discovery',
                default=False,
                help="Disabling iSCSI discovery (sendtargets) for multipath "
                     "connections on K2 driver."),
]

CONF = cfg.CONF
CONF.register_opts(kaminario_opts, group=configuration.SHARED_CONF_GROUP)

K2HTTPError = requests.exceptions.HTTPError
K2_RETRY_ERRORS = ("MC_ERR_BUSY", "MC_ERR_BUSY_SPECIFIC",
                   "MC_ERR_INPROGRESS", "MC_ERR_START_TIMEOUT")


class KaminarioCinderDriverException(exception.VolumeDriverException):
    message = _("KaminarioCinderDriver failure: %(reason)s")


class KaminarioRetryableException(exception.VolumeDriverException):
    message = _("Kaminario retryable exception: %(reason)s")


if krest:
    class KrestWrap(krest.EndPoint):
        def __init__(self, *args, **kwargs):
            self.krestlock = threading.Lock()
            super(KrestWrap, self).__init__(*args, **kwargs)

        def _should_retry(self, err_code, err_msg):
            if err_code == 400:
                for er in K2_RETRY_ERRORS:
                    if er in err_msg:
                        LOG.debug("Retry ERROR: %d with status %s",
                                  err_code, err_msg)
                        return True
            return False

        @utils.retry(KaminarioRetryableException,
                     retries=MAX_K2_RETRY)
        def _request(self, method, *args, **kwargs):
            try:
                self.krestlock.acquire()
                return super(KrestWrap, self)._request(method,
                                                       *args, **kwargs)
            except K2HTTPError as err:
                err_code = err.response.status_code
                err_msg = err.response.text
                if self._should_retry(err_code, err_msg):
                    raise KaminarioRetryableException(
                        reason=six.text_type(err_msg))
                raise
            finally:
                self.krestlock.release()


class Replication(object):
    def __init__(self, config, *args, **kwargs):
        self.backend_id = config.get('backend_id')
        self.login = config.get('login')
        self.password = config.get('password')
        self.rpo = config.get('rpo')


class KaminarioCinderDriver(cinder.volume.driver.ISCSIDriver):
    VENDOR = "Kaminario"
    stats = {}

    def __init__(self, *args, **kwargs):
        super(KaminarioCinderDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(san.san_opts)
        self.configuration.append_config_values(kaminario_opts)
        self.replica = None
        self._protocol = None
        k2_lock_sfx = self.configuration.safe_get('san_ip')
        self.k2_lock_name = "%s-%s" % (K2_LOCK_PREFIX, k2_lock_sfx)

    @classmethod
    def get_driver_options(cls):
        additional_opts = cls._get_oslo_driver_opts(
            'san_ip', 'san_login', 'san_password', 'replication_device',
            'volume_dd_blocksize', 'unique_fqdn_network')
        return kaminario_opts + additional_opts

    @volume_utils.trace
    def check_for_setup_error(self):
        if krest is None:
            msg = _("Unable to import 'krest' python module.")
            LOG.error(msg)
            raise KaminarioCinderDriverException(reason=msg)
        else:
            conf = self.configuration
            self.client = KrestWrap(conf.san_ip,
                                    conf.san_login,
                                    conf.san_password,
                                    ssl_validate=False)
            if self.replica:
                self.target = KrestWrap(self.replica.backend_id,
                                        self.replica.login,
                                        self.replica.password,
                                        ssl_validate=False)
            v_rs = self.client.search("system/state")
            if hasattr(v_rs, 'hits') and v_rs.total != 0:
                ver = v_rs.hits[0].rest_api_version
                ver_exist = versionutils.convert_version_to_int(ver)
                ver_min = versionutils.convert_version_to_int(K2_MIN_VERSION)
                if ver_exist < ver_min:
                    msg = _("K2 REST API version should be "
                            ">= %s.") % K2_MIN_VERSION
                    LOG.error(msg)
                    raise KaminarioCinderDriverException(reason=msg)

            else:
                msg = _("K2 REST API version search failed.")
                LOG.error(msg)
                raise KaminarioCinderDriverException(reason=msg)

    def _check_ops(self):
        """Ensure that the options we care about are set."""
        required_ops = ['san_ip', 'san_login', 'san_password']
        for attr in required_ops:
            if not getattr(self.configuration, attr, None):
                raise exception.InvalidInput(reason=_('%s is not set.') % attr)

        replica = self.configuration.safe_get('replication_device')
        if replica and isinstance(replica, list):
            replica_ops = ['backend_id', 'login', 'password', 'rpo']
            for attr in replica_ops:
                if attr not in replica[0]:
                    msg = _('replication_device %s is not set.') % attr
                    raise exception.InvalidInput(reason=msg)
            self.replica = Replication(replica[0])

    @volume_utils.trace
    def do_setup(self, context):
        super(KaminarioCinderDriver, self).do_setup(context)
        self._check_ops()

    @volume_utils.trace
    def create_volume(self, volume):
        """Volume creation in K2 needs a volume group.

        - create a volume group
        - create a volume in the volume group
        """
        vg_name = self.get_volume_group_name(volume.id)
        vol_name = self.get_volume_name(volume.id)
        prov_type = self._get_is_dedup(volume.get('volume_type'))
        try:
            LOG.debug("Creating volume group with name: %(name)s, "
                      "quota: unlimited and dedup_support: %(dedup)s",
                      {'name': vg_name, 'dedup': prov_type})

            vg = self.client.new("volume_groups", name=vg_name, quota=0,
                                 is_dedup=prov_type).save()
            LOG.debug("Creating volume with name: %(name)s, size: %(size)s "
                      "GB, volume_group: %(vg)s",
                      {'name': vol_name, 'size': volume.size, 'vg': vg_name})
            vol = self.client.new("volumes", name=vol_name,
                                  size=volume.size * units.Mi,
                                  volume_group=vg).save()
        except Exception as ex:
            vg_rs = self.client.search("volume_groups", name=vg_name)
            if vg_rs.total != 0:
                LOG.debug("Deleting vg: %s for failed volume in K2.", vg_name)
                vg_rs.hits[0].delete()
            LOG.exception("Creation of volume %s failed.", vol_name)
            raise KaminarioCinderDriverException(reason=ex)

        if self._get_is_replica(volume.volume_type) and self.replica:
            self._create_volume_replica(volume, vg, vol, self.replica.rpo)

    @volume_utils.trace
    def _create_volume_replica(self, volume, vg, vol, rpo):
        """Volume replica creation in K2 needs session and remote volume.

        - create a session
        - create a volume in the volume group

        """
        session_name = self.get_session_name(volume.id)
        rsession_name = self.get_rep_name(session_name)

        rvg_name = self.get_rep_name(vg.name)
        rvol_name = self.get_rep_name(vol.name)

        k2peer_rs = self.client.search("replication/peer_k2arrays",
                                       mgmt_host=self.replica.backend_id)
        if hasattr(k2peer_rs, 'hits') and k2peer_rs.total != 0:
            k2peer = k2peer_rs.hits[0]
        else:
            msg = _("Unable to find K2peer in source K2:")
            LOG.error(msg)
            raise KaminarioCinderDriverException(reason=msg)
        try:
            LOG.debug("Creating source session with name: %(sname)s and "
                      " target session name: %(tname)s",
                      {'sname': session_name, 'tname': rsession_name})
            src_ssn = self.client.new("replication/sessions")
            src_ssn.replication_peer_k2array = k2peer
            src_ssn.auto_configure_peer_volumes = "False"
            src_ssn.local_volume_group = vg
            src_ssn.replication_peer_volume_group_name = rvg_name
            src_ssn.remote_replication_session_name = rsession_name
            src_ssn.name = session_name
            src_ssn.rpo = rpo
            src_ssn.save()
            LOG.debug("Creating remote volume with name: %s",
                      rvol_name)
            self.client.new("replication/peer_volumes",
                            local_volume=vol,
                            name=rvol_name,
                            replication_session=src_ssn).save()
            src_ssn.state = "in_sync"
            src_ssn.save()
        except Exception as ex:
            LOG.exception("Replication for the volume %s has "
                          "failed.", vol.name)
            self._delete_by_ref(self.client, "replication/sessions",
                                session_name, 'session')
            self._delete_by_ref(self.target, "replication/sessions",
                                rsession_name, 'remote session')
            self._delete_by_ref(self.target, "volumes",
                                rvol_name, 'remote volume')
            self._delete_by_ref(self.client, "volumes", vol.name, "volume")
            self._delete_by_ref(self.target, "volume_groups",
                                rvg_name, "remote vg")
            self._delete_by_ref(self.client, "volume_groups", vg.name, "vg")
            raise KaminarioCinderDriverException(reason=ex)

    @volume_utils.trace
    def _create_failover_volume_replica(self, volume, vg_name, vol_name):
        """Volume replica creation in K2 needs session and remote volume.

        - create a session
        - create a volume in the volume group

        """
        session_name = self.get_session_name(volume.id)
        rsession_name = self.get_rep_name(session_name)

        rvg_name = self.get_rep_name(vg_name)
        rvol_name = self.get_rep_name(vol_name)
        rvg = self.target.search("volume_groups", name=rvg_name).hits[0]
        rvol = self.target.search("volumes", name=rvol_name).hits[0]
        k2peer_rs = self.target.search("replication/peer_k2arrays",
                                       mgmt_host=self.configuration.san_ip)
        if hasattr(k2peer_rs, 'hits') and k2peer_rs.total != 0:
            k2peer = k2peer_rs.hits[0]
        else:
            msg = _("Unable to find K2peer in source K2:")
            LOG.error(msg)
            raise KaminarioCinderDriverException(reason=msg)
        try:
            LOG.debug("Creating source session with name: %(sname)s and "
                      " target session name: %(tname)s",
                      {'sname': rsession_name, 'tname': session_name})
            tgt_ssn = self.target.new("replication/sessions")
            tgt_ssn.replication_peer_k2array = k2peer
            tgt_ssn.auto_configure_peer_volumes = "False"
            tgt_ssn.local_volume_group = rvg
            tgt_ssn.replication_peer_volume_group_name = vg_name
            tgt_ssn.remote_replication_session_name = session_name
            tgt_ssn.name = rsession_name
            tgt_ssn.rpo = self.replica.rpo
            tgt_ssn.save()
            LOG.debug("Creating remote volume with name: %s",
                      rvol_name)
            self.target.new("replication/peer_volumes",
                            local_volume=rvol,
                            name=vol_name,
                            replication_session=tgt_ssn).save()
            tgt_ssn.state = "in_sync"
            tgt_ssn.save()
        except Exception as ex:
            LOG.exception("Replication for the volume %s has "
                          "failed.", rvol_name)
            self._delete_by_ref(self.target, "replication/sessions",
                                rsession_name, 'session')
            self._delete_by_ref(self.client, "replication/sessions",
                                session_name, 'remote session')
            self._delete_by_ref(self.client, "volumes", vol_name, "volume")
            self._delete_by_ref(self.client, "volume_groups", vg_name, "vg")
            raise KaminarioCinderDriverException(reason=ex)

    @volume_utils.trace
    def _delete_by_ref(self, device, url, name, msg):
        rs = device.search(url, name=name)
        for result in rs.hits:
            result.delete()
            LOG.debug("Deleting %(msg)s: %(name)s", {'msg': msg, 'name': name})

    @volume_utils.trace
    def _failover_volume(self, volume):
        """Promoting a secondary volume to primary volume."""
        session_name = self.get_session_name(volume.id)
        rsession_name = self.get_rep_name(session_name)
        tgt_ssn = self.target.search("replication/sessions",
                                     name=rsession_name).hits[0]
        if tgt_ssn.state == 'in_sync':
            tgt_ssn.state = 'failed_over'
            tgt_ssn.save()
            LOG.debug("The target session: %s state is "
                      "changed to failed_over ", rsession_name)

    @volume_utils.trace
    def failover_host(self, context, volumes, secondary_id=None, groups=None):
        """Failover to replication target."""
        volume_updates = []
        back_end_ip = None
        svc_host = volume_utils.extract_host(self.host, 'backend')
        service = objects.Service.get_by_args(context, svc_host,
                                              'cinder-volume')

        if secondary_id and secondary_id != self.replica.backend_id:
            LOG.error("Kaminario driver received failover_host "
                      "request, But backend is non replicated device")
            raise exception.UnableToFailOver(reason=_("Failover requested "
                                                      "on non replicated "
                                                      "backend."))

        if (service.active_backend_id and
                service.active_backend_id != self.configuration.san_ip):
            self.snap_updates = []
            rep_volumes = []
            # update status for non-replicated primary volumes
            for v in volumes:
                vol_name = self.get_volume_name(v['id'])
                vol = self.client.search("volumes", name=vol_name)
                if v.replication_status != K2_REP_FAILED_OVER and vol.total:
                    status = 'available'
                    if v.volume_attachment:
                        map_rs = self.client.search("mappings",
                                                    volume=vol.hits[0])
                        status = 'in-use'
                        if map_rs.total:
                            map_rs.hits[0].delete()
                    volume_updates.append({'volume_id': v['id'],
                                           'updates':
                                           {'status': status}})
                else:
                    rep_volumes.append(v)

            # In-sync from secondaray array to primary array
            for v in rep_volumes:
                vol_name = self.get_volume_name(v['id'])
                vol = self.client.search("volumes", name=vol_name)
                rvol_name = self.get_rep_name(vol_name)
                rvol = self.target.search("volumes", name=rvol_name)
                session_name = self.get_session_name(v['id'])
                rsession_name = self.get_rep_name(session_name)
                ssn = self.target.search("replication/sessions",
                                         name=rsession_name)
                if ssn.total:
                    tgt_ssn = ssn.hits[0]
                ssn = self.client.search("replication/sessions",
                                         name=session_name)
                if ssn.total:
                    src_ssn = ssn.hits[0]

                if (tgt_ssn.state == 'failed_over' and
                   tgt_ssn.current_role == 'target' and vol.total and src_ssn):
                    map_rs = self.client.search("mappings", volume=vol.hits[0])
                    if map_rs.total:
                        map_rs.hits[0].delete()
                    tgt_ssn.state = 'in_sync'
                    tgt_ssn.save()
                    self._check_for_status(src_ssn, 'in_sync')
                if (rvol.total and src_ssn.state == 'in_sync' and
                   src_ssn.current_role == 'target'):
                    gen_no = self._create_volume_replica_user_snap(self.target,
                                                                   tgt_ssn)
                    self.snap_updates.append({'tgt_ssn': tgt_ssn,
                                              'gno': gen_no,
                                              'stime': time.time()})
                LOG.debug("The target session: %s state is "
                          "changed to in sync", rsession_name)

            self._is_user_snap_sync_finished()

            # Delete secondary volume mappings and create snapshot
            for v in rep_volumes:
                vol_name = self.get_volume_name(v['id'])
                vol = self.client.search("volumes", name=vol_name)
                rvol_name = self.get_rep_name(vol_name)
                rvol = self.target.search("volumes", name=rvol_name)
                session_name = self.get_session_name(v['id'])
                rsession_name = self.get_rep_name(session_name)
                ssn = self.target.search("replication/sessions",
                                         name=rsession_name)
                if ssn.total:
                    tgt_ssn = ssn.hits[0]
                ssn = self.client.search("replication/sessions",
                                         name=session_name)
                if ssn.total:
                    src_ssn = ssn.hits[0]
                if (rvol.total and src_ssn.state == 'in_sync' and
                   src_ssn.current_role == 'target'):
                    map_rs = self.target.search("mappings",
                                                volume=rvol.hits[0])
                    if map_rs.total:
                        map_rs.hits[0].delete()
                    gen_no = self._create_volume_replica_user_snap(self.target,
                                                                   tgt_ssn)
                    self.snap_updates.append({'tgt_ssn': tgt_ssn,
                                              'gno': gen_no,
                                              'stime': time.time()})
            self._is_user_snap_sync_finished()
            # changing source sessions to failed-over
            for v in rep_volumes:
                vol_name = self.get_volume_name(v['id'])
                vol = self.client.search("volumes", name=vol_name)
                rvol_name = self.get_rep_name(vol_name)
                rvol = self.target.search("volumes", name=rvol_name)
                session_name = self.get_session_name(v['id'])
                rsession_name = self.get_rep_name(session_name)
                ssn = self.target.search("replication/sessions",
                                         name=rsession_name)
                if ssn.total:
                    tgt_ssn = ssn.hits[0]
                ssn = self.client.search("replication/sessions",
                                         name=session_name)
                if ssn.total:
                    src_ssn = ssn.hits[0]
                if (rvol.total and src_ssn.state == 'in_sync' and
                   src_ssn.current_role == 'target'):
                    src_ssn.state = 'failed_over'
                    src_ssn.save()
                    self._check_for_status(tgt_ssn, 'suspended')
                    LOG.debug("The target session: %s state is "
                              "changed to failed over", session_name)

                    src_ssn.state = 'in_sync'
                    src_ssn.save()
                    LOG.debug("The target session: %s state is "
                              "changed to in sync", session_name)
                    rep_status = fields.ReplicationStatus.DISABLED
                    volume_updates.append({'volume_id': v['id'],
                                           'updates':
                                          {'replication_status': rep_status}})

            back_end_ip = self.configuration.san_ip
        else:
            """Failover to replication target."""
            for v in volumes:
                vol_name = self.get_volume_name(v['id'])
                rv = self.get_rep_name(vol_name)
                if self.target.search("volumes", name=rv).total:
                    self._failover_volume(v)
                    volume_updates.append(
                        {'volume_id': v['id'],
                         'updates':
                         {'replication_status': K2_REP_FAILED_OVER}})
                else:
                    volume_updates.append({'volume_id': v['id'],
                                           'updates': {'status': 'error', }})
            back_end_ip = self.replica.backend_id
        return back_end_ip, volume_updates, []

    @volume_utils.trace
    def _create_volume_replica_user_snap(self, k2, sess):
        snap = k2.new("snapshots")
        snap.is_application_consistent = "False"
        snap.replication_session = sess
        snap.save()
        return snap.generation_number

    def _is_user_snap_sync_finished(self):
        # waiting for user snapshot to be synced
        while len(self.snap_updates) > 0:
            for update in self.snap_updates:
                sess = update.get('tgt_ssn')
                gno = update.get('gno')
                stime = update.get('stime')
                sess.refresh()
                if (sess.generation_number == gno and
                   sess.current_snapshot_progress == 100
                   and sess.current_snapshot_id is None):
                    if time.time() - stime > 300:
                        gen_no = self._create_volume_replica_user_snap(
                            self.target,
                            sess)
                        self.snap_updates.append({'tgt_ssn': sess,
                                                  'gno': gen_no,
                                                  'stime': time.time()})
                    self.snap_updates.remove(update)
                eventlet.sleep(1)

    @volume_utils.trace
    def create_volume_from_snapshot(self, volume, snapshot):
        """Create volume from snapshot.

        - search for snapshot and retention_policy
        - create a view from snapshot and attach view
        - create a volume and attach volume
        - copy data from attached view to attached volume
        - detach volume and view and finally delete view
        """
        snap_name = self.get_snap_name(snapshot.id)
        view_name = self.get_view_name(volume.id)
        vol_name = self.get_volume_name(volume.id)
        cview = src_attach_info = dest_attach_info = None
        rpolicy = self.get_policy()
        properties = volume_utils.brick_get_connector_properties()
        LOG.debug("Searching for snapshot: %s in K2.", snap_name)
        snap_rs = self.client.search("snapshots", short_name=snap_name)
        if hasattr(snap_rs, 'hits') and snap_rs.total != 0:
            snap = snap_rs.hits[0]
            LOG.debug("Creating a view: %(view)s from snapshot: %(snap)s",
                      {'view': view_name, 'snap': snap_name})
            try:
                cview = self.client.new("snapshots",
                                        short_name=view_name,
                                        source=snap, retention_policy=rpolicy,
                                        is_exposable=True).save()
            except Exception as ex:
                LOG.exception("Creating a view: %(view)s from snapshot: "
                              "%(snap)s failed", {"view": view_name,
                                                  "snap": snap_name})
                raise KaminarioCinderDriverException(reason=ex)

        else:
            msg = _("Snapshot: %s search failed in K2.") % snap_name
            LOG.error(msg)
            raise KaminarioCinderDriverException(reason=msg)

        try:
            conn = self.initialize_connection(cview, properties)
            src_attach_info = self._connect_device(conn)
            self.create_volume(volume)
            conn = self.initialize_connection(volume, properties)
            dest_attach_info = self._connect_device(conn)
            volume_utils.copy_volume(src_attach_info['device']['path'],
                                     dest_attach_info['device']['path'],
                                     snapshot.volume.size * units.Ki,
                                     self.configuration.volume_dd_blocksize,
                                     sparse=True)
            self._kaminario_disconnect_volume(src_attach_info,
                                              dest_attach_info)
            self.terminate_connection(volume, properties)
            self.terminate_connection(cview, properties)
            cview.delete()
        except Exception as ex:
            self._kaminario_disconnect_volume(src_attach_info,
                                              dest_attach_info)
            self.terminate_connection(cview, properties)
            self.terminate_connection(volume, properties)
            cview.delete()
            self.delete_volume(volume)
            LOG.exception("Copy to volume: %(vol)s from view: %(view)s "
                          "failed", {"vol": vol_name, "view": view_name})
            raise KaminarioCinderDriverException(reason=ex)

    @volume_utils.trace
    def create_cloned_volume(self, volume, src_vref):
        """Create a clone from source volume.

        - attach source volume
        - create and attach new volume
        - copy data from attached source volume to attached new volume
        - detach both volumes
        """
        clone_name = self.get_volume_name(volume.id)
        src_name = self.get_volume_name(src_vref.id)
        src_vol = self.client.search("volumes", name=src_name)
        src_map = self.client.search("mappings", volume=src_vol)
        src_attach_info = dest_attach_info = None
        if src_map.total != 0:
            msg = _("K2 driver does not support clone of an attached volume. "
                    "To get this done, create a snapshot from the attached "
                    "volume and then create a volume from the snapshot.")
            LOG.error(msg)
            raise KaminarioCinderDriverException(reason=msg)
        try:
            properties = volume_utils.brick_get_connector_properties()
            conn = self.initialize_connection(src_vref, properties)
            src_attach_info = self._connect_device(conn)
            self.create_volume(volume)
            conn = self.initialize_connection(volume, properties)
            dest_attach_info = self._connect_device(conn)
            volume_utils.copy_volume(src_attach_info['device']['path'],
                                     dest_attach_info['device']['path'],
                                     src_vref.size * units.Ki,
                                     self.configuration.volume_dd_blocksize,
                                     sparse=True)
            self._kaminario_disconnect_volume(src_attach_info,
                                              dest_attach_info)
            self.terminate_connection(volume, properties)
            self.terminate_connection(src_vref, properties)
        except Exception as ex:
            self._kaminario_disconnect_volume(src_attach_info,
                                              dest_attach_info)
            self.terminate_connection(src_vref, properties)
            self.terminate_connection(volume, properties)
            self.delete_volume(volume)
            LOG.exception("Create a clone: %s failed.", clone_name)
            raise KaminarioCinderDriverException(reason=ex)

    @volume_utils.trace
    def delete_volume(self, volume):
        """Volume in K2 exists in a volume group.

        - delete the volume
        - delete the corresponding volume group
        """
        vg_name = self.get_volume_group_name(volume.id)
        vol_name = self.get_volume_name(volume.id)
        try:
            if self._get_is_replica(volume.volume_type) and self.replica:
                self._delete_volume_replica(volume, vg_name, vol_name)

            LOG.debug("Searching and deleting volume: %s in K2.", vol_name)
            vol_rs = self.client.search("volumes", name=vol_name)
            if vol_rs.total != 0:
                vol_rs.hits[0].delete()
            LOG.debug("Searching and deleting vg: %s in K2.", vg_name)
            vg_rs = self.client.search("volume_groups", name=vg_name)
            if vg_rs.total != 0:
                vg_rs.hits[0].delete()
        except Exception as ex:
            LOG.exception("Deletion of volume %s failed.", vol_name)
            raise KaminarioCinderDriverException(reason=ex)

    @volume_utils.trace
    def _delete_volume_replica(self, volume, vg_name, vol_name):
        rvg_name = self.get_rep_name(vg_name)
        rvol_name = self.get_rep_name(vol_name)
        session_name = self.get_session_name(volume.id)
        rsession_name = self.get_rep_name(session_name)
        src_ssn = self.client.search('replication/sessions',
                                     name=session_name).hits[0]
        tgt_ssn = self.target.search('replication/sessions',
                                     name=rsession_name).hits[0]
        src_ssn.state = 'suspended'
        src_ssn.save()
        self._check_for_status(tgt_ssn, 'suspended')
        src_ssn.state = 'idle'
        src_ssn.save()
        self._check_for_status(tgt_ssn, 'idle')
        tgt_ssn.delete()
        src_ssn.delete()

        LOG.debug("Searching and deleting snapshots for volume groups:"
                  "%(vg1)s, %(vg2)s in K2.", {'vg1': vg_name, 'vg2': rvg_name})
        vg = self.client.search('volume_groups', name=vg_name).hits
        rvg = self.target.search('volume_groups', name=rvg_name).hits
        snaps = self.client.search('snapshots', volume_group=vg).hits
        for s in snaps:
            s.delete()
        rsnaps = self.target.search('snapshots', volume_group=rvg).hits
        for s in rsnaps:
            s.delete()

        self._delete_by_ref(self.target, "volumes", rvol_name, 'remote volume')
        self._delete_by_ref(self.target, "volume_groups",
                            rvg_name, "remote vg")

    @volume_utils.trace
    def _delete_failover_volume_replica(self, volume, vg_name, vol_name):
        rvg_name = self.get_rep_name(vg_name)
        rvol_name = self.get_rep_name(vol_name)
        session_name = self.get_session_name(volume.id)
        rsession_name = self.get_rep_name(session_name)
        tgt_ssn = self.target.search('replication/sessions',
                                     name=rsession_name).hits[0]
        tgt_ssn.state = 'idle'
        tgt_ssn.save()
        tgt_ssn.delete()

        LOG.debug("Searching and deleting snapshots for target volume group "
                  "and target volume: %(vol)s, %(vg)s in K2.",
                  {'vol': rvol_name, 'vg': rvg_name})
        rvg = self.target.search('volume_groups', name=rvg_name).hits
        rsnaps = self.target.search('snapshots', volume_group=rvg).hits
        for s in rsnaps:
            s.delete()

    def _check_for_status(self, obj, status):
        while obj.state != status:
            obj.refresh()
            eventlet.sleep(1)

    @volume_utils.trace
    def get_volume_stats(self, refresh=False):
        if refresh:
            self.update_volume_stats()
        stats = self.stats
        stats['storage_protocol'] = self._protocol
        stats['driver_version'] = self.VERSION
        stats['vendor_name'] = self.VENDOR
        backend_name = self.configuration.safe_get('volume_backend_name')
        stats['volume_backend_name'] = (backend_name or
                                        self.__class__.__name__)
        return stats

    def create_export(self, context, volume, connector):
        pass

    def ensure_export(self, context, volume):
        pass

    def remove_export(self, context, volume):
        pass

    @volume_utils.trace
    def create_snapshot(self, snapshot):
        """Create a snapshot from a volume_group."""
        vg_name = self.get_volume_group_name(snapshot.volume_id)
        snap_name = self.get_snap_name(snapshot.id)
        rpolicy = self.get_policy()
        try:
            LOG.debug("Searching volume_group: %s in K2.", vg_name)
            vg = self.client.search("volume_groups", name=vg_name).hits[0]
            LOG.debug("Creating a snapshot: %(snap)s from vg: %(vg)s",
                      {'snap': snap_name, 'vg': vg_name})
            self.client.new("snapshots", short_name=snap_name,
                            source=vg, retention_policy=rpolicy,
                            is_auto_deleteable=False).save()
        except Exception as ex:
            LOG.exception("Creation of snapshot: %s failed.", snap_name)
            raise KaminarioCinderDriverException(reason=ex)

    @volume_utils.trace
    def delete_snapshot(self, snapshot):
        """Delete a snapshot."""
        snap_name = self.get_snap_name(snapshot.id)
        try:
            LOG.debug("Searching and deleting snapshot: %s in K2.", snap_name)
            snap_rs = self.client.search("snapshots", short_name=snap_name)
            if snap_rs.total != 0:
                snap_rs.hits[0].delete()
        except Exception as ex:
            LOG.exception("Deletion of snapshot: %s failed.", snap_name)
            raise KaminarioCinderDriverException(reason=ex)

    @volume_utils.trace
    def extend_volume(self, volume, new_size):
        """Extend volume."""
        vol_name = self.get_volume_name(volume.id)
        try:
            LOG.debug("Searching volume: %s in K2.", vol_name)
            vol = self.client.search("volumes", name=vol_name).hits[0]
            vol.size = new_size * units.Mi
            LOG.debug("Extending volume: %s in K2.", vol_name)
            vol.save()
        except Exception as ex:
            LOG.exception("Extending volume: %s failed.", vol_name)
            raise KaminarioCinderDriverException(reason=ex)

    def update_volume_stats(self):
        conf = self.configuration
        LOG.debug("Searching system capacity in K2.")
        cap = self.client.search("system/capacity").hits[0]
        LOG.debug("Searching total volumes in K2 for updating stats.")
        total_volumes = self.client.search("volumes").total - 1
        provisioned_vol = cap.provisioned_volumes

        if (conf.auto_calc_max_oversubscription_ratio and cap.provisioned
                and (cap.total - cap.free) != 0):
            ratio = provisioned_vol / float(cap.total - cap.free)
        else:
            ratio = volume_utils.get_max_over_subscription_ratio(
                conf.max_over_subscription_ratio, supports_auto=True)

        self.stats = {'QoS_support': False,
                      'free_capacity_gb': cap.free / units.Mi,
                      'total_capacity_gb': cap.total / units.Mi,
                      'thin_provisioning_support': True,
                      'sparse_copy_volume': True,
                      'total_volumes': total_volumes,
                      'thick_provisioning_support': False,
                      'provisioned_capacity_gb': provisioned_vol / units.Mi,
                      'max_over_subscription_ratio': ratio,
                      'kaminario:thin_prov_type': 'dedup/nodedup',
                      'replication_enabled': True,
                      'kaminario:replication': True}

    def get_initiator_host_name(self, connector):
        """Return the initiator host name or unique ID.

        Unique ID when configuration's unique_fqdn_network is false will be
        the reversed IQN/WWPNS.

        Valid characters: 0-9, a-z, A-Z, '-', '_'
        All other characters are replaced with '_'.
        Total characters in initiator host name: 32
        """
        name = connector.get('initiator',
                             connector.get('wwnns', [''])[0])[::-1]
        if self.configuration.unique_fqdn_network:
            name = connector.get('host', name)
        return re.sub('[^0-9a-zA-Z-_]', '_', name[:32])

    def get_volume_group_name(self, vid):
        """Return the volume group name."""
        return "cvg-{0}".format(vid)

    def get_volume_name(self, vid):
        """Return the volume name."""
        return "cv-{0}".format(vid)

    def get_session_name(self, vid):
        """Return the volume name."""
        return "ssn-{0}".format(vid)

    def get_snap_name(self, sid):
        """Return the snapshot name."""
        return "cs-{0}".format(sid)

    def get_view_name(self, vid):
        """Return the view name."""
        return "cview-{0}".format(vid)

    def get_rep_name(self, name):
        """Return the corresponding replication names."""
        return "r{0}".format(name)

    @volume_utils.trace
    def _delete_host_by_name(self, name):
        """Deleting host by name."""
        host_rs = self.client.search("hosts", name=name)
        if hasattr(host_rs, "hits") and host_rs.total != 0:
            host = host_rs.hits[0]
            host.delete()

    def get_policy(self):
        """Return the retention policy."""
        try:
            LOG.debug("Searching for retention_policy in K2.")
            return self.client.search("retention_policies",
                                      name="Best_Effort_Retention").hits[0]
        except Exception as ex:
            LOG.exception("Retention policy search failed in K2.")
            raise KaminarioCinderDriverException(reason=ex)

    def _get_volume_object(self, volume):
        vol_name = self.get_volume_name(volume.id)
        if volume.replication_status == K2_REP_FAILED_OVER:
            vol_name = self.get_rep_name(vol_name)
        LOG.debug("Searching volume : %s in K2.", vol_name)
        vol_rs = self.client.search("volumes", name=vol_name)
        if not hasattr(vol_rs, 'hits') or vol_rs.total == 0:
            msg = _("Unable to find volume: %s from K2.") % vol_name
            LOG.error(msg)
            raise KaminarioCinderDriverException(reason=msg)
        return vol_rs.hits[0]

    def _get_lun_number(self, vol, host):
        volsnap = None
        LOG.debug("Searching volsnaps in K2.")
        volsnap_rs = self.client.search("volsnaps", snapshot=vol)
        if hasattr(volsnap_rs, 'hits') and volsnap_rs.total != 0:
            volsnap = volsnap_rs.hits[0]

        LOG.debug("Searching mapping of volsnap in K2.")
        map_rs = self.client.search("mappings", volume=volsnap, host=host)
        return map_rs.hits[0].lun

    def initialize_connection(self, volume, connector):
        pass

    @volume_utils.trace
    def terminate_connection(self, volume, connector):
        """Terminate connection of volume from host."""
        # Get volume object
        if type(volume).__name__ != 'RestObject':
            vol_name = self.get_volume_name(volume.id)
            if volume.replication_status == K2_REP_FAILED_OVER:
                vol_name = self.get_rep_name(vol_name)
            LOG.debug("Searching volume: %s in K2.", vol_name)
            volume_rs = self.client.search("volumes", name=vol_name)
            if hasattr(volume_rs, "hits") and volume_rs.total != 0:
                volume = volume_rs.hits[0]
        else:
            vol_name = volume.name
        host_name = ""
        if connector is None:
            vol_map_rs = self.client.search("mappings", {"volume": volume})
            if hasattr(vol_map_rs, "hits") and vol_map_rs.total != 0:
                host_name = vol_map_rs.hits[0].host.name
        else:
            # Get host object.
            host_name = self.get_initiator_host_name(connector)
        host_rs = self.client.search("hosts", name=host_name)
        if hasattr(host_rs, "hits") and host_rs.total != 0 and volume:
            host = host_rs.hits[0]
            LOG.debug("Searching and deleting mapping of volume: %(name)s to "
                      "host: %(host)s", {'host': host_name, 'name': vol_name})
            map_rs = self.client.search("mappings", volume=volume, host=host)
            if hasattr(map_rs, "hits") and map_rs.total != 0:
                map_rs.hits[0].delete()
            if self.client.search("mappings", host=host).total == 0:
                LOG.debug("Deleting initiator hostname: %s in K2.", host_name)
                host.delete()
        else:
            LOG.warning("Host: %s not found on K2.", host_name)

    @volume_utils.trace
    def k2_initialize_connection(self, volume, connector):
        # Get volume object.
        if type(volume).__name__ != 'RestObject':
            vol = self._get_volume_object(volume)
        else:
            vol = volume
        # Get host object.
        host, host_rs, host_name = self._get_host_object(connector)
        try:
            # Map volume object to host object.
            LOG.debug("Mapping volume: %(vol)s to host: %(host)s",
                      {'host': host_name, 'vol': vol.name})
            mapping = self.client.new("mappings", volume=vol, host=host).save()
        except Exception as ex:
            if host_rs.total == 0:
                self._delete_host_by_name(host_name)
            LOG.exception("Unable to map volume: %(vol)s to host: "
                          "%(host)s", {'host': host_name,
                                       'vol': vol.name})
            raise KaminarioCinderDriverException(reason=ex)
        # Get lun number.
        if type(volume).__name__ == 'RestObject':
            return self._get_lun_number(vol, host)
        else:
            return mapping.lun

    def _get_host_object(self, connector):
        pass

    def _get_is_dedup(self, vol_type):
        if vol_type:
            specs_val = vol_type.get('extra_specs', {}).get(
                'kaminario:thin_prov_type')
            if specs_val == 'nodedup':
                return False
            else:
                return True
        else:
            return True

    def _get_is_replica(self, vol_type):
        replica = False
        if vol_type and vol_type.get('extra_specs'):
            specs = vol_type.get('extra_specs')
            if (specs.get('kaminario:replication') == 'enabled' and
               self.replica):
                replica = True
        return replica

    def _get_replica_status(self, vg_name):
        vg_rs = self.client.search("volume_groups", name=vg_name)
        if vg_rs.total:
            vg = vg_rs.hits[0]
            if self.client.search("replication/sessions",
                                  local_volume_group=vg).total:
                return True
        return False

    @volume_utils.trace
    def manage_existing(self, volume, existing_ref):
        vol_name = existing_ref['source-name']
        new_name = self.get_volume_name(volume.id)
        vg_new_name = self.get_volume_group_name(volume.id)
        vg_name = None
        is_dedup = self._get_is_dedup(volume.get('volume_type'))
        reason = None
        try:
            LOG.debug("Searching volume: %s in K2.", vol_name)
            vol = self.client.search("volumes", name=vol_name).hits[0]
            vg = vol.volume_group
            nvol = self.client.search("volumes", volume_group=vg).total
            vg_replica = self._get_replica_status(vg.name)
            vol_map = False
            if self.client.search("mappings", volume=vol).total != 0:
                vol_map = True
            if is_dedup != vg.is_dedup:
                reason = 'dedup type mismatch for K2 volume group.'
            elif vg_replica:
                reason = 'replication enabled K2 volume group.'
            elif vol_map:
                reason = 'attached K2 volume.'
            elif nvol != 1:
                reason = 'multiple volumes in K2 volume group.'
            if reason:
                raise exception.ManageExistingInvalidReference(
                    existing_ref=existing_ref,
                    reason=_('Unable to manage K2 volume due to: %s') % reason)
            vol.name = new_name
            vg_name = vg.name
            LOG.debug("Manage new volume name: %s", new_name)
            vg.name = vg_new_name
            LOG.debug("Manage volume group name: %s", vg_new_name)
            vg.save()
            LOG.debug("Manage volume: %s in K2.", vol_name)
            vol.save()
        except exception.ManageExistingInvalidReference:
            LOG.exception("manage volume: %s failed.", vol_name)
            raise
        except Exception:
            LOG.exception("manage volume: %s failed.", vol_name)
            vg_rs = self.client.search("volume_groups", name=vg_new_name)
            if hasattr(vg_rs, 'hits') and vg_rs.total != 0:
                vg = vg_rs.hits[0]
                if vg_name and vg.name == vg_new_name:
                    vg.name = vg_name
                    LOG.debug("Updating vg new name to old name: %s ", vg_name)
                    vg.save()
            raise

    @volume_utils.trace
    def manage_existing_get_size(self, volume, existing_ref):
        vol_name = existing_ref['source-name']
        v_rs = self.client.search("volumes", name=vol_name)
        if hasattr(v_rs, 'hits') and v_rs.total != 0:
            vol = v_rs.hits[0]
            size = vol.size / units.Mi
            return math.ceil(size)
        else:
            raise exception.ManageExistingInvalidReference(
                existing_ref=existing_ref,
                reason=_('Unable to get size of manage volume.'))

    @volume_utils.trace
    def after_volume_copy(self, ctxt, volume, new_volume, remote=None):
        self.delete_volume(volume)
        vg_name_old = self.get_volume_group_name(volume.id)
        vol_name_old = self.get_volume_name(volume.id)
        vg_name_new = self.get_volume_group_name(new_volume.id)
        vol_name_new = self.get_volume_name(new_volume.id)
        vg_new = self.client.search("volume_groups", name=vg_name_new).hits[0]
        vg_new.name = vg_name_old
        vg_new.save()
        vol_new = self.client.search("volumes", name=vol_name_new).hits[0]
        vol_new.name = vol_name_old
        vol_new.save()

    @volume_utils.trace
    def retype(self, ctxt, volume, new_type, diff, host):
        old_type = volume.get('volume_type')
        vg_name = self.get_volume_group_name(volume.id)
        vol_name = self.get_volume_name(volume.id)
        vol_rs = self.client.search("volumes", name=vol_name)
        if vol_rs.total:
            vol = vol_rs.hits[0]
            vmap = self.client.search("mappings", volume=vol).total
        old_rep_type = self._get_replica_status(vg_name)
        new_rep_type = self._get_is_replica(new_type)
        new_prov_type = self._get_is_dedup(new_type)
        old_prov_type = self._get_is_dedup(old_type)
        # Change dedup<->nodedup with add/remove replication is complex in K2
        # since K2 does not have api to change dedup<->nodedup.
        if new_prov_type == old_prov_type:
            if not old_rep_type and new_rep_type:
                self._add_replication(volume)
                return True
            elif old_rep_type and not new_rep_type:
                self._delete_replication(volume)
                return True
        elif not new_rep_type and not old_rep_type:
            msg = ("Use '--migration-policy on-demand' to change 'dedup "
                   "without replication'<->'nodedup without replication'.")
            if vol_rs.total and vmap:
                msg = "Unattach volume and {0}".format(msg)
            LOG.debug(msg)
            return False
        else:
            LOG.error('Change from type1: %(type1)s to type2: %(type2)s '
                      'is not supported directly in K2.',
                      {'type1': old_type, 'type2': new_type})
            return False

    def _add_replication(self, volume):
        vg_name = self.get_volume_group_name(volume.id)
        vol_name = self.get_volume_name(volume.id)
        if volume.replication_status == K2_REP_FAILED_OVER:
            self._create_failover_volume_replica(volume, vg_name, vol_name)
        else:
            LOG.debug("Searching volume group with name: %(name)s",
                      {'name': vg_name})
            vg = self.client.search("volume_groups", name=vg_name).hits[0]
            LOG.debug("Searching volume with name: %(name)s",
                      {'name': vol_name})
            vol = self.client.search("volumes", name=vol_name).hits[0]
            self._create_volume_replica(volume, vg, vol, self.replica.rpo)

    def _delete_replication(self, volume):
        vg_name = self.get_volume_group_name(volume.id)
        vol_name = self.get_volume_name(volume.id)
        if volume.replication_status == K2_REP_FAILED_OVER:
            self._delete_failover_volume_replica(volume, vg_name, vol_name)
        else:
            self._delete_volume_replica(volume, vg_name, vol_name)

    def _kaminario_disconnect_volume(self, *attach_info):
        for info in attach_info:
            if (info and info.get('connector') and
                    info.get('conn', {}).get('data') and info.get('device')):
                info['connector'].disconnect_volume(info['conn']['data'],
                                                    info['device'])



# Copyright 2021, Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Tests for User Facing Messages in Backup Operations."""

from unittest import mock

from cinder.backup import manager as backup_manager
from cinder import exception
from cinder.message import message_field
from cinder.scheduler import manager as sch_manager
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test


class BackupUserMessagesTest(test.TestCase):

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager._run_backup')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    def test_backup_create_invalid_status(
            self, mock_notify, mock_working, mock_run,
            mock_msg_create, mock_get_vol, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='available', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = {'status': 'backing-up'}.__getitem__
        mock_get_vol.return_value = mock_vol

        self.assertRaises(
            exception.InvalidBackup, manager.create_backup, fake_context,
            fake_backup)
        self.assertEqual(message_field.Action.BACKUP_CREATE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_INVALID_STATE)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager._run_backup')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    def test_backup_create_service_down(
            self, mock_notify, mock_working, mock_run, mock_msg_create,
            mock_get_vol, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = {'status': 'backing-up'}.__getitem__
        mock_get_vol.return_value = mock_vol
        mock_working.return_value = False

        mock_run.side_effect = exception.InvalidBackup(reason='test reason')
        self.assertRaises(
            exception.InvalidBackup, manager.create_backup, fake_context,
            fake_backup)
        self.assertEqual(message_field.Action.BACKUP_CREATE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_SERVICE_DOWN)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    @mock.patch(
        'cinder.backup.manager.volume_utils.brick_get_connector_properties')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.get_backup_device')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_cleanup_temp_volumes_snapshots_when_backup_created')
    def test_backup_create_device_error(
            self, mock_cleanup, mock_get_bak_dev, mock_get_conn, mock_notify,
            mock_working, mock_msg_create, mock_get_vol, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = {'status': 'backing-up'}.__getitem__
        mock_get_vol.return_value = mock_vol
        mock_working.return_value = True
        mock_get_bak_dev.side_effect = exception.InvalidVolume(
            reason="test reason")

        self.assertRaises(exception.InvalidVolume, manager.create_backup,
                          fake_context, fake_backup)
        self.assertEqual(message_field.Action.BACKUP_CREATE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_CREATE_DEVICE_ERROR)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    @mock.patch(
        'cinder.backup.manager.volume_utils.brick_get_connector_properties')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.get_backup_device')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_cleanup_temp_volumes_snapshots_when_backup_created')
    @mock.patch('cinder.backup.manager.BackupManager._attach_device')
    def test_backup_create_attach_error(
            self, mock_attach, mock_cleanup, mock_get_bak_dev, mock_get_conn,
            mock_notify, mock_working, mock_msg_create, mock_get_vol,
            mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = {'status': 'backing-up'}.__getitem__
        mock_get_vol.return_value = mock_vol
        mock_working.return_value = True
        mock_attach.side_effect = exception.InvalidVolume(reason="test reason")

        self.assertRaises(exception.InvalidVolume, manager.create_backup,
                          fake_context, fake_backup)
        self.assertEqual(message_field.Action.BACKUP_CREATE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.ATTACH_ERROR)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    @mock.patch(
        'cinder.backup.manager.volume_utils.brick_get_connector_properties')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.get_backup_device')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_cleanup_temp_volumes_snapshots_when_backup_created')
    @mock.patch('cinder.backup.manager.BackupManager._attach_device')
    @mock.patch(
        'cinder.tests.unit.backup.fake_service.FakeBackupService.backup')
    @mock.patch('cinder.backup.manager.open')
    @mock.patch('cinder.backup.manager.BackupManager._detach_device')
    def test_backup_create_driver_error(
            self, mock_detach, mock_open, mock_backup, mock_attach,
            mock_cleanup, mock_get_bak_dev, mock_get_conn, mock_notify,
            mock_working, mock_msg_create, mock_get_vol, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = {'status': 'backing-up'}.__getitem__
        mock_get_vol.return_value = mock_vol
        mock_working.return_value = True
        mock_attach.return_value = {'device': {'path': '/dev/sdb'}}
        mock_backup.side_effect = exception.InvalidBackup(reason="test reason")

        self.assertRaises(exception.InvalidBackup, manager.create_backup,
                          fake_context, fake_backup)
        self.assertEqual(message_field.Action.BACKUP_CREATE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_CREATE_DRIVER_ERROR)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    @mock.patch(
        'cinder.backup.manager.volume_utils.brick_get_connector_properties')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.get_backup_device')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_cleanup_temp_volumes_snapshots_when_backup_created')
    @mock.patch('cinder.backup.manager.BackupManager._attach_device')
    @mock.patch(
        'cinder.tests.unit.backup.fake_service.FakeBackupService.backup')
    @mock.patch('cinder.backup.manager.open')
    @mock.patch('cinder.backup.manager.BackupManager._detach_device')
    def test_backup_create_detach_error(
            self, mock_detach, mock_open, mock_backup, mock_attach,
            mock_cleanup, mock_get_bak_dev, mock_get_conn, mock_notify,
            mock_working, mock_msg_create, mock_get_vol, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = {'status': 'backing-up'}.__getitem__
        mock_get_vol.return_value = mock_vol
        mock_working.return_value = True
        mock_attach.return_value = {'device': {'path': '/dev/sdb'}}
        mock_detach.side_effect = exception.InvalidVolume(reason="test reason")

        self.assertRaises(exception.InvalidVolume, manager.create_backup,
                          fake_context, fake_backup)
        self.assertEqual(message_field.Action.BACKUP_CREATE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.DETACH_ERROR)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    @mock.patch(
        'cinder.backup.manager.volume_utils.brick_get_connector_properties')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.get_backup_device')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_cleanup_temp_volumes_snapshots_when_backup_created')
    @mock.patch('cinder.backup.manager.BackupManager._attach_device')
    @mock.patch(
        'cinder.tests.unit.backup.fake_service.FakeBackupService.backup')
    @mock.patch('cinder.backup.manager.open')
    @mock.patch('cinder.backup.manager.BackupManager._detach_device')
    def test_backup_create_cleanup_error(
            self, mock_detach, mock_open, mock_backup, mock_attach,
            mock_cleanup, mock_get_bak_dev, mock_get_conn, mock_notify,
            mock_working, mock_msg_create, mock_get_vol, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = {'status': 'backing-up'}.__getitem__
        mock_get_vol.return_value = mock_vol
        mock_working.return_value = True
        mock_attach.return_value = {'device': {'path': '/dev/sdb'}}
        mock_cleanup.side_effect = exception.InvalidVolume(
            reason="test reason")

        self.assertRaises(exception.InvalidVolume, manager.create_backup,
                          fake_context, fake_backup)
        self.assertEqual(message_field.Action.BACKUP_CREATE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_CREATE_CLEANUP_ERROR)

    @mock.patch('cinder.scheduler.host_manager.HostManager.'
                '_get_available_backup_service_host')
    @mock.patch('cinder.volume.volume_utils.update_backup_error')
    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.db.volume_get')
    @mock.patch('cinder.message.api.API.create')
    def test_backup_create_scheduling_error(
            self, mock_msg_create, mock_get_vol, mock_vol_update,
            mock_update_error, mock_get_backup_host):
        manager = sch_manager.SchedulerManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(id=fake.BACKUP_ID,
                                     volume_id=fake.VOLUME_ID)
        mock_get_vol.return_value = mock.MagicMock()
        exception.ServiceNotFound(service_id='cinder-backup')
        mock_get_backup_host.side_effect = exception.ServiceNotFound(
            service_id='cinder-backup')

        manager.create_backup(fake_context, fake_backup)
        mock_msg_create.assert_called_once_with(
            fake_context,
            action=message_field.Action.BACKUP_CREATE,
            resource_type=message_field.Resource.VOLUME_BACKUP,
            resource_uuid=fake_backup.id,
            detail=message_field.Detail.BACKUP_SCHEDULE_ERROR)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch(
        'cinder.backup.manager.BackupManager._notify_about_backup_usage')
    def test_backup_delete_invalid_state(
            self, mock_notify, mock_msg_create, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='available', volume_id=fake.VOLUME_ID,
            snapshot_id=None)

        self.assertRaises(
            exception.InvalidBackup, manager.delete_backup, fake_context,
            fake_backup)
        self.assertEqual(message_field.Action.BACKUP_DELETE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_INVALID_STATE)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch(
        'cinder.backup.manager.BackupManager._notify_about_backup_usage')
    def test_backup_delete_service_down(
            self, mock_notify, mock_working, mock_msg_create,
            mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='deleting', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        mock_working.return_value = False

        self.assertRaises(
            exception.InvalidBackup, manager.delete_backup, fake_context,
            fake_backup)
        self.assertEqual(message_field.Action.BACKUP_DELETE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_SERVICE_DOWN)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager._is_our_backup')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch(
        'cinder.backup.manager.BackupManager._notify_about_backup_usage')
    def test_backup_delete_driver_error(
            self, mock_notify, mock_working, mock_our_back,
            mock_msg_create, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='deleting', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        fake_backup.__getitem__.side_effect = (
            {'display_name': 'fail_on_delete'}.__getitem__)
        mock_working.return_value = True
        mock_our_back.return_value = True

        self.assertRaises(
            IOError, manager.delete_backup, fake_context,
            fake_backup)
        self.assertEqual(message_field.Action.BACKUP_DELETE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_DELETE_DRIVER_ERROR)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    def test_backup_restore_volume_invalid_state(
            self, mock_notify, mock_msg_create, mock_get_vol,
            mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        fake_backup.__getitem__.side_effect = (
            {'status': 'restoring', 'size': 1}.__getitem__)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = (
            {'id': fake.VOLUME_ID, 'status': 'available',
             'size': 1}.__getitem__)
        mock_get_vol.return_value = mock_vol

        self.assertRaises(
            exception.InvalidVolume, manager.restore_backup,
            fake_context, fake_backup, fake.VOLUME_ID)
        mock_msg_create.assert_called_once_with(
            fake_context,
            action=message_field.Action.BACKUP_RESTORE,
            resource_type=message_field.Resource.VOLUME_BACKUP,
            resource_uuid=mock_vol.id,
            detail=message_field.Detail.VOLUME_INVALID_STATE)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    def test_backup_restore_backup_invalid_state(
            self, mock_notify, mock_msg_create, mock_get_vol,
            mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        fake_backup.__getitem__.side_effect = (
            {'status': 'available', 'size': 1}.__getitem__)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = (
            {'status': 'restoring-backup', 'size': 1}.__getitem__)
        mock_get_vol.return_value = mock_vol

        self.assertRaises(
            exception.InvalidBackup, manager.restore_backup,
            fake_context, fake_backup, fake.VOLUME_ID)
        self.assertEqual(message_field.Action.BACKUP_RESTORE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_INVALID_STATE)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager._is_our_backup')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    @mock.patch(
        'cinder.backup.manager.volume_utils.brick_get_connector_properties')
    @mock.patch(
        'cinder.volume.rpcapi.VolumeAPI.secure_file_operations_enabled')
    @mock.patch('cinder.backup.manager.BackupManager._attach_device')
    @mock.patch('cinder.backup.manager.BackupManager._detach_device')
    def test_backup_restore_attach_error(
            self, mock_detach, mock_attach, mock_sec_opts, mock_get_conn,
            mock_notify, mock_working, mock_our_back, mock_msg_create,
            mock_get_vol, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        fake_backup.__getitem__.side_effect = (
            {'status': 'restoring', 'size': 1}.__getitem__)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = (
            {'status': 'restoring-backup', 'size': 1}.__getitem__)
        mock_get_vol.return_value = mock_vol
        mock_working.return_value = True
        mock_our_back.return_value = True
        mock_attach.side_effect = exception.InvalidBackup(
            reason="test reason")

        self.assertRaises(
            exception.InvalidBackup, manager.restore_backup,
            fake_context, fake_backup, fake.VOLUME_ID)
        self.assertEqual(message_field.Action.BACKUP_RESTORE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.ATTACH_ERROR)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager._is_our_backup')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    @mock.patch(
        'cinder.backup.manager.volume_utils.brick_get_connector_properties')
    @mock.patch(
        'cinder.volume.rpcapi.VolumeAPI.secure_file_operations_enabled')
    @mock.patch('cinder.backup.manager.BackupManager._attach_device')
    @mock.patch('cinder.backup.manager.open')
    @mock.patch(
        'cinder.tests.unit.backup.fake_service.FakeBackupService.restore')
    @mock.patch('cinder.backup.manager.BackupManager._detach_device')
    def test_backup_restore_driver_error(
            self, mock_detach, mock_restore, mock_open, mock_attach,
            mock_sec_opts, mock_get_conn, mock_notify, mock_working,
            mock_our_back, mock_msg_create, mock_get_vol, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        fake_backup.__getitem__.side_effect = (
            {'status': 'restoring', 'size': 1}.__getitem__)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = (
            {'status': 'restoring-backup', 'size': 1}.__getitem__)
        mock_get_vol.return_value = mock_vol
        mock_working.return_value = True
        mock_our_back.return_value = True
        mock_attach.return_value = {'device': {'path': '/dev/sdb'}}
        mock_restore.side_effect = exception.InvalidBackup(
            reason="test reason")

        self.assertRaises(
            exception.InvalidBackup, manager.restore_backup,
            fake_context, fake_backup, fake.VOLUME_ID)
        self.assertEqual(message_field.Action.BACKUP_RESTORE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.BACKUP_RESTORE_ERROR)

    @mock.patch('cinder.db.volume_update')
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.message.api.API.create_from_request_context')
    @mock.patch('cinder.backup.manager.BackupManager._is_our_backup')
    @mock.patch('cinder.backup.manager.BackupManager.is_working')
    @mock.patch('cinder.backup.manager.BackupManager.'
                '_notify_about_backup_usage')
    @mock.patch(
        'cinder.backup.manager.volume_utils.brick_get_connector_properties')
    @mock.patch(
        'cinder.volume.rpcapi.VolumeAPI.secure_file_operations_enabled')
    @mock.patch('cinder.backup.manager.BackupManager._attach_device')
    @mock.patch('cinder.backup.manager.open')
    @mock.patch(
        'cinder.tests.unit.backup.fake_service.FakeBackupService.restore')
    @mock.patch('cinder.backup.manager.BackupManager._detach_device')
    def test_backup_restore_detach_error(
            self, mock_detach, mock_restore, mock_open, mock_attach,
            mock_sec_opts, mock_get_conn, mock_notify, mock_working,
            mock_our_back, mock_msg_create, mock_get_vol, mock_vol_update):
        manager = backup_manager.BackupManager()
        fake_context = mock.MagicMock()
        fake_backup = mock.MagicMock(
            id=fake.BACKUP_ID, status='creating', volume_id=fake.VOLUME_ID,
            snapshot_id=None)
        fake_backup.__getitem__.side_effect = (
            {'status': 'restoring', 'size': 1}.__getitem__)
        mock_vol = mock.MagicMock()
        mock_vol.__getitem__.side_effect = (
            {'status': 'restoring-backup', 'size': 1}.__getitem__)
        mock_get_vol.return_value = mock_vol
        mock_working.return_value = True
        mock_our_back.return_value = True
        mock_attach.return_value = {'device': {'path': '/dev/sdb'}}
        mock_detach.side_effect = exception.InvalidBackup(
            reason="test reason")

        self.assertRaises(
            exception.InvalidBackup, manager.restore_backup,
            fake_context, fake_backup, fake.VOLUME_ID)
        self.assertEqual(message_field.Action.BACKUP_RESTORE,
                         fake_context.message_action)
        self.assertEqual(message_field.Resource.VOLUME_BACKUP,
                         fake_context.message_resource_type)
        self.assertEqual(fake_backup.id,
                         fake_context.message_resource_id)
        mock_msg_create.assert_called_with(
            fake_context,
            detail=message_field.Detail.DETACH_ERROR)



# Copyright (c) 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Tests for Posix backup driver."""

import builtins
import os
from unittest import mock

from cinder.backup.drivers import posix
from cinder import context
from cinder import objects
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test


FAKE_FILE_SIZE = 52428800
FAKE_SHA_BLOCK_SIZE_BYTES = 1024
FAKE_BACKUP_ENABLE_PROGRESS_TIMER = True

FAKE_CONTAINER = 'fake/container'
FAKE_BACKUP_ID = fake.BACKUP_ID
FAKE_BACKUP_ID_PART1 = fake.BACKUP_ID[:2]
FAKE_BACKUP_ID_PART2 = fake.BACKUP_ID[2:4]
FAKE_BACKUP_ID_REST = fake.BACKUP_ID[4:]
FAKE_BACKUP = {'id': FAKE_BACKUP_ID, 'container': None}

UPDATED_CONTAINER_NAME = os.path.join(FAKE_BACKUP_ID_PART1,
                                      FAKE_BACKUP_ID_PART2,
                                      FAKE_BACKUP_ID)

FAKE_BACKUP_MOUNT_POINT_BASE = '/fake/mount-point-base'
FAKE_EXPORT_PATH = 'fake/export/path'

FAKE_BACKUP_POSIX_PATH = os.path.join(FAKE_BACKUP_MOUNT_POINT_BASE,
                                      FAKE_EXPORT_PATH)

FAKE_PREFIX = 'prefix-'
FAKE_CONTAINER_ENTRIES = [FAKE_PREFIX + 'one', FAKE_PREFIX + 'two', 'three']
EXPECTED_CONTAINER_ENTRIES = [FAKE_PREFIX + 'one', FAKE_PREFIX + 'two']
FAKE_OBJECT_NAME = 'fake-object-name'
FAKE_OBJECT_PATH = os.path.join(FAKE_BACKUP_POSIX_PATH, FAKE_CONTAINER,
                                FAKE_OBJECT_NAME)


class PosixBackupDriverTestCase(test.TestCase):

    def setUp(self):
        super(PosixBackupDriverTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

        self.override_config('backup_file_size',
                             FAKE_FILE_SIZE)
        self.override_config('backup_sha_block_size_bytes',
                             FAKE_SHA_BLOCK_SIZE_BYTES)
        self.override_config('backup_enable_progress_timer',
                             FAKE_BACKUP_ENABLE_PROGRESS_TIMER)
        self.override_config('backup_posix_path',
                             FAKE_BACKUP_POSIX_PATH)
        self.mock_object(posix, 'LOG')

        self.driver = posix.PosixBackupDriver(self.ctxt)

    def test_init(self):
        drv = posix.PosixBackupDriver(self.ctxt)
        self.assertEqual(FAKE_BACKUP_POSIX_PATH,
                         drv.backup_path)

    def test_update_container_name_container_passed(self):
        result = self.driver.update_container_name(FAKE_BACKUP, FAKE_CONTAINER)

        self.assertEqual(FAKE_CONTAINER, result)

    def test_update_container_na_container_passed(self):
        result = self.driver.update_container_name(FAKE_BACKUP, None)

        self.assertEqual(UPDATED_CONTAINER_NAME, result)

    def test_put_container(self):
        self.mock_object(os.path, 'exists', return_value=False)
        self.mock_object(os, 'makedirs')
        self.mock_object(os, 'chmod')
        path = os.path.join(self.driver.backup_path, FAKE_CONTAINER)

        self.driver.put_container(FAKE_CONTAINER)

        os.path.exists.assert_called_once_with(path)
        os.makedirs.assert_called_once_with(path)
        os.chmod.assert_called_once_with(path, 0o770)

    def test_put_container_already_exists(self):
        self.mock_object(os.path, 'exists', return_value=True)
        self.mock_object(os, 'makedirs')
        self.mock_object(os, 'chmod')
        path = os.path.join(self.driver.backup_path, FAKE_CONTAINER)

        self.driver.put_container(FAKE_CONTAINER)

        os.path.exists.assert_called_once_with(path)
        self.assertEqual(0, os.makedirs.call_count)
        self.assertEqual(0, os.chmod.call_count)

    def test_put_container_exception(self):
        self.mock_object(os.path, 'exists', return_value=False)
        self.mock_object(os, 'makedirs', side_effect=OSError)
        self.mock_object(os, 'chmod')
        path = os.path.join(self.driver.backup_path, FAKE_CONTAINER)

        self.assertRaises(OSError, self.driver.put_container,
                          FAKE_CONTAINER)
        os.path.exists.assert_called_once_with(path)
        os.makedirs.assert_called_once_with(path)
        self.assertEqual(0, os.chmod.call_count)

    def test_get_container_entries(self):
        self.mock_object(os, 'listdir', return_value=FAKE_CONTAINER_ENTRIES)

        result = self.driver.get_container_entries(FAKE_CONTAINER, FAKE_PREFIX)

        self.assertEqual(EXPECTED_CONTAINER_ENTRIES, result)

    def test_get_container_entries_no_list(self):
        self.mock_object(os, 'listdir', return_value=[])

        result = self.driver.get_container_entries(FAKE_CONTAINER, FAKE_PREFIX)

        self.assertEqual([], result)

    def test_get_container_entries_no_match(self):
        self.mock_object(os, 'listdir', return_value=FAKE_CONTAINER_ENTRIES)

        result = self.driver.get_container_entries(FAKE_CONTAINER,
                                                   FAKE_PREFIX + 'garbage')

        self.assertEqual([], result)

    def test_get_object_writer(self):
        self.mock_object(builtins, 'open', mock.mock_open())
        self.mock_object(os, 'chmod')

        self.driver.get_object_writer(FAKE_CONTAINER, FAKE_OBJECT_NAME)

        os.chmod.assert_called_once_with(FAKE_OBJECT_PATH, 0o660)
        builtins.open.assert_called_once_with(FAKE_OBJECT_PATH, 'wb')

    def test_get_object_reader(self):
        self.mock_object(builtins, 'open', mock.mock_open())

        self.driver.get_object_reader(FAKE_CONTAINER, FAKE_OBJECT_NAME)

        builtins.open.assert_called_once_with(FAKE_OBJECT_PATH, 'rb')

    def test_delete_object(self):
        self.mock_object(os, 'remove')

        self.driver.delete_object(FAKE_CONTAINER, FAKE_OBJECT_NAME)

    @mock.patch.object(posix.timeutils, 'utcnow')
    def test_generate_object_name_prefix(self, utcnow_mock):
        timestamp = '20170518102205'
        utcnow_mock.return_value.strftime.return_value = timestamp
        backup = objects.Backup(self.ctxt, volume_id=fake.VOLUME_ID,
                                id=fake.BACKUP_ID)
        res = self.driver._generate_object_name_prefix(backup)
        expected = 'volume_%s_%s_backup_%s' % (backup.volume_id,
                                               timestamp,
                                               backup.id)
        self.assertEqual(expected, res)



#    (c) Copyright 2013-2015 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Volume driver for HPE 3PAR Storage array.

This driver requires 3.1.3 or later firmware on the 3PAR array, using
the 4.x version of the hpe3parclient.

You will need to install the python hpe3parclient.
sudo pip install --upgrade "hpe3parclient>=4.0"

Set the following in the cinder.conf file to enable the
3PAR Fibre Channel Driver along with the required flags:

volume_driver=cinder.volume.drivers.hpe.hpe_3par_fc.HPE3PARFCDriver
"""

try:
    from hpe3parclient import exceptions as hpeexceptions
except ImportError:
    hpeexceptions = None

from oslo_log import log as logging
from oslo_utils.excutils import save_and_reraise_exception

from cinder import coordination
from cinder import interface
from cinder.volume.drivers.hpe import hpe_3par_base as hpebasedriver
from cinder.volume import volume_utils
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)

# EXISTENT_PATH error code returned from hpe3parclient
EXISTENT_PATH = 73


@interface.volumedriver
class HPE3PARFCDriver(hpebasedriver.HPE3PARDriverBase):
    """OpenStack Fibre Channel driver to enable 3PAR storage array.

    Version history:

    .. code-block:: none

        1.0   - Initial driver
        1.1   - QoS, extend volume, multiple iscsi ports, remove domain,
                session changes, faster clone, requires 3.1.2 MU2 firmware,
                copy volume <--> Image.
        1.2.0 - Updated the use of the hp3parclient to 2.0.0 and refactored
                the drivers to use the new APIs.
        1.2.1 - Synchronized extend_volume method.
        1.2.2 - Added try/finally around client login/logout.
        1.2.3 - Added ability to add WWNs to host.
        1.2.4 - Added metadata during attach/detach bug #1258033.
        1.3.0 - Removed all SSH code.  We rely on the hp3parclient now.
        2.0.0 - Update hp3parclient API uses 3.0.x
        2.0.2 - Add back-end assisted volume migrate
        2.0.3 - Added initiator-target map for FC Zone Manager
        2.0.4 - Added support for managing/unmanaging of volumes
        2.0.5 - Only remove FC Zone on last volume detach
        2.0.6 - Added support for volume retype
        2.0.7 - Only one FC port is used when a single FC path
                is present.  bug #1360001
        2.0.8 - Fixing missing login/logout around attach/detach bug #1367429
        2.0.9 - Add support for pools with model update
        2.0.10 - Migrate without losing type settings bug #1356608
        2.0.11 - Removing locks bug #1381190
        2.0.12 - Fix queryHost call to specify wwns bug #1398206
        2.0.13 - Fix missing host name during attach bug #1398206
        2.0.14 - Removed usage of host name cache #1398914
        2.0.15 - Added support for updated detach_volume attachment.
        2.0.16 - Added encrypted property to initialize_connection #1439917
        2.0.17 - Improved VLUN creation and deletion logic. #1469816
        2.0.18 - Changed initialize_connection to use getHostVLUNs. #1475064
        2.0.19 - Adds consistency group support
        2.0.20 - Update driver to use ABC metaclasses
        2.0.21 - Added update_migrated_volume. bug # 1492023
        3.0.0 - Rebranded HP to HPE.
        3.0.1 - Remove db access for consistency groups
        3.0.2 - Adds v2 managed replication support
        3.0.3 - Adds v2 unmanaged replication support
        3.0.4 - Adding manage/unmanage snapshot support
        3.0.5 - Optimize array ID retrieval
        3.0.6 - Update replication to version 2.1
        3.0.7 - Remove metadata that tracks the instance ID. bug #1572665
        3.0.8 - NSP feature, creating FC Vlun as match set instead of
                host sees. bug #1577993
        3.0.9 - Handling HTTP conflict 409, host WWN/iSCSI name already used
                by another host, while creating 3PAR FC Host. bug #1597454
        3.0.10 - Added Entry point tracing
        3.0.11 - Handle manage and unmanage hosts present. bug #1648067
        3.0.12 - Adds consistency group capability in generic volume groups.
        4.0.0 - Adds base class.
        4.0.1 - Added check to remove FC zones. bug #1730720
        4.0.2 - Create one vlun in single path configuration. bug #1727176
        4.0.3 - Create FC vlun as host sees. bug #1734505
        4.0.4 - Handle force detach case. bug #1686745
        4.0.5 - Set proper backend on subsequent operation, after group
                failover. bug #1773069
        4.0.6 - Set NSP for single path attachments. Bug #1809249
        4.0.7 - Added Peer Persistence feature

    """

    VERSION = "4.0.7"

    # The name of the CI wiki page.
    CI_WIKI_NAME = "HPE_Storage_CI"

    def __init__(self, *args, **kwargs):
        super(HPE3PARFCDriver, self).__init__(*args, **kwargs)
        self.lookup_service = fczm_utils.create_lookup_service()
        self.protocol = 'FC'

    def _initialize_connection_common(self, volume, connector, common, host,
                                      target_wwns, init_targ_map, numPaths,
                                      remote_client=None):
        # check if a VLUN already exists for this host
        existing_vlun = common.find_existing_vlun(volume, host, remote_client)

        vlun = None
        if existing_vlun is None:
            # now that we have a host, create the VLUN
            if self.lookup_service and numPaths == 1:
                nsp = None
                active_fc_port_list = (
                    common.get_active_fc_target_ports(remote_client))
                for port in active_fc_port_list:
                    if port['portWWN'].lower() == target_wwns[0].lower():
                        nsp = port['nsp']
                        break
                vlun = common.create_vlun(volume, host, nsp, None,
                                          remote_client)
            else:
                vlun = common.create_vlun(volume, host, None, None,
                                          remote_client)
        else:
            vlun = existing_vlun

        info_backend = {'driver_volume_type': 'fibre_channel',
                        'data': {'target_lun': vlun['lun'],
                                 'target_discovered': True,
                                 'target_wwn': target_wwns,
                                 'initiator_target_map': init_targ_map}}

        encryption_key_id = volume.get('encryption_key_id')
        info_backend['data']['encrypted'] = encryption_key_id is not None
        fczm_utils.add_fc_zone(info_backend)

        return info_backend

    @volume_utils.trace
    @coordination.synchronized('3par-{volume.id}')
    def initialize_connection(self, volume, connector):
        """Assigns the volume to a server.

        Assign any created volume to a compute node/host so that it can be
        used from that host.

        The  driver returns a driver_volume_type of 'fibre_channel'.
        The target_wwn can be a single entry or a list of wwns that
        correspond to the list of remote wwn(s) that will export the volume.
        Example return values:

            {
                'driver_volume_type': 'fibre_channel'
                'data': {
                    'encrypted': False,
                    'target_discovered': True,
                    'target_lun': 1,
                    'target_wwn': '1234567890123',
                }
            }

            or

             {
                'driver_volume_type': 'fibre_channel'
                'data': {
                    'encrypted': False,
                    'target_discovered': True,
                    'target_lun': 1,
                    'target_wwn': ['1234567890123', '0987654321321'],
                }
            }


        Steps to export a volume on 3PAR
          * Create a host on the 3par with the target wwn
          * Create a VLUN for that HOST with the volume we want to export.

        """
        LOG.debug("volume id: %(volume_id)s",
                  {'volume_id': volume['id']})
        array_id = self.get_volume_replication_driver_data(volume)
        common = self._login(array_id=array_id)
        try:
            # we have to make sure we have a host
            host, cpg = self._create_host(common, volume, connector)
            target_wwns, init_targ_map, numPaths = (
                self._build_initiator_target_map(common, connector))

            multipath = connector.get('multipath')
            LOG.debug("multipath: %(multipath)s",
                      {'multipath': multipath})

            user_target = None
            if not multipath:
                user_target = self._get_user_target(common)
                initiator = connector.get('wwpns')[0]
                if user_target is None:
                    target_wwns = target_wwns[:1]
                    init_targ_map[initiator] = init_targ_map[initiator][:1]
                else:
                    target_wwns = [user_target]
                    init_targ_map[initiator] = [user_target]

            info = self._initialize_connection_common(
                volume, connector, common, host,
                target_wwns, init_targ_map, numPaths)

            if not multipath:
                return info

            if volume.get('replication_status') != 'enabled':
                return info

            LOG.debug('This is a replication setup')

            remote_target = common._replication_targets[0]
            replication_mode = remote_target['replication_mode']
            quorum_witness_ip = remote_target.get('quorum_witness_ip')

            if replication_mode == 1:
                LOG.debug('replication_mode is sync')
                if quorum_witness_ip:
                    LOG.debug('quorum_witness_ip is present')
                    LOG.debug('Peer Persistence has been configured')
                else:
                    LOG.debug('Since quorum_witness_ip is absent, '
                              'considering this as Active/Passive '
                              'replication')
                    return info
            else:
                LOG.debug('Active/Passive replication has been '
                          'configured')
                return info

            # Peer Persistence has been configured
            remote_client = common._create_replication_client(remote_target)

            host, cpg = self._create_host(
                common, volume, connector,
                remote_target, cpg, remote_client)
            target_wwns, init_targ_map, numPaths = (
                self._build_initiator_target_map(
                    common, connector, remote_client))

            info_peer = self._initialize_connection_common(
                volume, connector, common, host,
                target_wwns, init_targ_map, numPaths,
                remote_client)

            common._destroy_replication_client(remote_client)

            info = {'driver_volume_type': 'fibre_channel',
                    'data': {'encrypted': info['data']['encrypted'],
                             'target_lun': info['data']['target_lun'],
                             'target_discovered': True,
                             'target_wwn': info['data']['target_wwn'] +
                             info_peer['data']['target_wwn'],
                             'initiator_target_map': self.merge_dicts(
                             info['data']['initiator_target_map'],
                             info_peer['data']['initiator_target_map'])}}

            return info
        finally:
            self._logout(common)

    @volume_utils.trace
    @coordination.synchronized('3par-{volume.id}')
    def terminate_connection(self, volume, connector, **kwargs):
        """Driver entry point to detach a volume from an instance."""
        array_id = self.get_volume_replication_driver_data(volume)
        common = self._login(array_id=array_id)
        try:
            is_force_detach = connector is None

            remote_client = None
            multipath = False
            if connector:
                multipath = connector.get('multipath')
            LOG.debug("multipath: %(multipath)s",
                      {'multipath': multipath})
            if multipath:
                if volume.get('replication_status') == 'enabled':
                    LOG.debug('This is a replication setup')

                    remote_target = common._replication_targets[0]
                    replication_mode = remote_target['replication_mode']
                    quorum_witness_ip = (
                        remote_target.get('quorum_witness_ip'))

                    if replication_mode == 1:
                        LOG.debug('replication_mode is sync')
                        if quorum_witness_ip:
                            LOG.debug('quorum_witness_ip is present')
                            LOG.debug('Peer Persistence has been configured')
                        else:
                            LOG.debug('Since quorum_witness_ip is absent, '
                                      'considering this as Active/Passive '
                                      'replication')
                    else:
                        LOG.debug('Active/Passive replication has been '
                                  'configured')

                    if replication_mode == 1 and quorum_witness_ip:
                        remote_client = (
                            common._create_replication_client(remote_target))

            if is_force_detach:
                common.terminate_connection(volume, None, None)
                # TODO(sonivi): remove zones, if not required
                # for now, do not remove zones
                zone_remove = False
            else:
                hostname = common._safe_hostname(connector, self.configuration)
                common.terminate_connection(volume, hostname,
                                            wwn=connector['wwpns'],
                                            remote_client=remote_client)

                zone_remove = True
                try:
                    vluns = common.client.getHostVLUNs(hostname)
                except hpeexceptions.HTTPNotFound:
                    # No more exports for this host.
                    pass
                else:
                    # Vlun exists, so check for wwpn entry.
                    for wwpn in connector.get('wwpns'):
                        for vlun in vluns:
                            if (vlun.get('active') and
                                    vlun.get('remoteName') == wwpn.upper()):
                                zone_remove = False
                                break

            info = {'driver_volume_type': 'fibre_channel',
                    'data': {}}

            if zone_remove:
                LOG.info("Need to remove FC Zone, building initiator "
                         "target map")
                target_wwns, init_targ_map, _numPaths = (
                    self._build_initiator_target_map(common, connector))

                info['data'] = {'target_wwn': target_wwns,
                                'initiator_target_map': init_targ_map}
                fczm_utils.remove_fc_zone(info)

            if remote_client:
                if zone_remove:
                    try:
                        vluns = remote_client.getHostVLUNs(hostname)
                    except hpeexceptions.HTTPNotFound:
                        # No more exports for this host.
                        pass
                    else:
                        # Vlun exists, so check for wwpn entry.
                        for wwpn in connector.get('wwpns'):
                            for vlun in vluns:
                                if (vlun.get('active') and
                                   vlun.get('remoteName') == wwpn.upper()):
                                    zone_remove = False
                                    break

                info_peer = {'driver_volume_type': 'fibre_channel',
                             'data': {}}

                if zone_remove:
                    LOG.info("Need to remove FC Zone, building initiator "
                             "target map")
                    target_wwns, init_targ_map, _numPaths = (
                        self._build_initiator_target_map(common, connector,
                                                         remote_client))

                    info_peer['data'] = {'target_wwn': target_wwns,
                                         'initiator_target_map': init_targ_map}
                    fczm_utils.remove_fc_zone(info_peer)

                    info = (
                        {'driver_volume_type': 'fibre_channel',
                         'data': {'target_wwn': info['data']['target_wwn'] +
                                  info_peer['data']['target_wwn'],
                                  'initiator_target_map': self.merge_dicts(
                                  info['data']['initiator_target_map'],
                                  info_peer['data']['initiator_target_map'])}})

            return info

        finally:
            self._logout(common)

    def _build_initiator_target_map(self, common, connector,
                                    remote_client=None):
        """Build the target_wwns and the initiator target map."""

        fc_ports = common.get_active_fc_target_ports(remote_client)
        all_target_wwns = []
        target_wwns = []
        init_targ_map = {}
        numPaths = 0

        for port in fc_ports:
            all_target_wwns.append(port['portWWN'])

        if self.lookup_service is not None:
            # use FC san lookup to determine which NSPs to use
            # for the new VLUN.
            dev_map = self.lookup_service.get_device_mapping_from_network(
                connector['wwpns'],
                all_target_wwns)

            for fabric_name in dev_map:
                fabric = dev_map[fabric_name]
                target_wwns += fabric['target_port_wwn_list']
                for initiator in fabric['initiator_port_wwn_list']:
                    if initiator not in init_targ_map:
                        init_targ_map[initiator] = []
                    init_targ_map[initiator] += fabric['target_port_wwn_list']
                    init_targ_map[initiator] = list(set(
                        init_targ_map[initiator]))
                    for _target in init_targ_map[initiator]:
                        numPaths += 1
            target_wwns = list(set(target_wwns))
        else:
            initiator_wwns = connector['wwpns']
            target_wwns = all_target_wwns

            for initiator in initiator_wwns:
                init_targ_map[initiator] = target_wwns

        return target_wwns, init_targ_map, numPaths

    def _create_3par_fibrechan_host(self, common, hostname, wwns,
                                    domain, persona_id, remote_client=None):
        """Create a 3PAR host.

        Create a 3PAR host, if there is already a host on the 3par using
        the same wwn but with a different hostname, return the hostname
        used by 3PAR.
        """
        # first search for an existing host
        host_found = None

        if remote_client:
            client_obj = remote_client
        else:
            client_obj = common.client

        hosts = client_obj.queryHost(wwns=wwns)

        if hosts and hosts['members'] and 'name' in hosts['members'][0]:
            host_found = hosts['members'][0]['name']

        if host_found is not None:
            return host_found
        else:
            persona_id = int(persona_id)
            try:
                client_obj.createHost(hostname, FCWwns=wwns,
                                      optional={'domain': domain,
                                                'persona': persona_id})
            except hpeexceptions.HTTPConflict as path_conflict:
                msg = "Create FC host caught HTTP conflict code: %s"
                LOG.exception(msg, path_conflict.get_code())
                with save_and_reraise_exception(reraise=False) as ctxt:
                    if path_conflict.get_code() is EXISTENT_PATH:
                        # Handle exception : EXISTENT_PATH - host WWN/iSCSI
                        # name already used by another host
                        hosts = client_obj.queryHost(wwns=wwns)
                        if hosts and hosts['members'] and (
                                'name' in hosts['members'][0]):
                            hostname = hosts['members'][0]['name']
                        else:
                            # re rasise last caught exception
                            ctxt.reraise = True
                    else:
                        # re rasise last caught exception
                        # for other HTTP conflict
                        ctxt.reraise = True
            return hostname

    def _modify_3par_fibrechan_host(self, common, hostname, wwn,
                                    remote_client):
        if remote_client:
            client_obj = remote_client
        else:
            client_obj = common.client

        mod_request = {'pathOperation': client_obj.HOST_EDIT_ADD,
                       'FCWWNs': wwn}
        try:
            client_obj.modifyHost(hostname, mod_request)
        except hpeexceptions.HTTPConflict as path_conflict:
            msg = ("Modify FC Host %(hostname)s caught "
                   "HTTP conflict code: %(code)s")
            LOG.exception(msg,
                          {'hostname': hostname,
                           'code': path_conflict.get_code()})

    def _create_host(self, common, volume, connector,
                     remote_target=None, src_cpg=None, remote_client=None):
        """Creates or modifies existing 3PAR host."""
        host = None
        domain = None
        hostname = common._safe_hostname(connector, self.configuration)
        if remote_target:
            cpg = common._get_cpg_from_cpg_map(
                remote_target['cpg_map'], src_cpg)
            cpg_obj = remote_client.getCPG(cpg)
            if 'domain' in cpg_obj:
                domain = cpg_obj['domain']
        else:
            cpg = common.get_cpg(volume, allowSnap=True)
            domain = common.get_domain(cpg)

        if not connector.get('multipath'):
            connector['wwpns'] = connector['wwpns'][:1]
        try:
            if remote_target:
                host = remote_client.getHost(hostname)
            else:
                host = common._get_3par_host(hostname)
                # Check whether host with wwn of initiator present on 3par
                hosts = common.client.queryHost(wwns=connector['wwpns'])
                host, hostname = (
                    common._get_prioritized_host_on_3par(
                        host, hosts, hostname))
        except hpeexceptions.HTTPNotFound:
            # get persona from the volume type extra specs
            persona_id = common.get_persona_type(volume)
            # host doesn't exist, we have to create it
            hostname = self._create_3par_fibrechan_host(common,
                                                        hostname,
                                                        connector['wwpns'],
                                                        domain,
                                                        persona_id,
                                                        remote_client)
            if remote_target:
                host = remote_client.getHost(hostname)
            else:
                host = common._get_3par_host(hostname)
            return host, cpg
        else:
            host = self._add_new_wwn_to_host(
                common, host, connector['wwpns'], remote_client)
            return host, cpg

    def _add_new_wwn_to_host(self, common, host, wwns, remote_client=None):
        """Add wwns to a host if one or more don't exist.

        Identify if argument wwns contains any world wide names
        not configured in the 3PAR host path. If any are found,
        add them to the 3PAR host.
        """
        # get the currently configured wwns
        # from the host's FC paths
        host_wwns = []
        if 'FCPaths' in host:
            for path in host['FCPaths']:
                wwn = path.get('wwn', None)
                if wwn is not None:
                    host_wwns.append(wwn.lower())

        # lower case all wwns in the compare list
        compare_wwns = [x.lower() for x in wwns]

        # calculate wwns in compare list, but not in host_wwns list
        new_wwns = list(set(compare_wwns).difference(host_wwns))

        # if any wwns found that were not in host list,
        # add them to the host
        if (len(new_wwns) > 0):
            self._modify_3par_fibrechan_host(
                common, host['name'], new_wwns, remote_client)
            if remote_client:
                host = remote_client.getHost(host['name'])
            else:
                host = common._get_3par_host(host['name'])
        return host

    def _get_user_target(self, common):
        target_nsp = common.config.hpe3par_target_nsp

        if not target_nsp:
            return None

        # Get target wwn from target nsp
        fc_ports = common.get_active_fc_target_ports()

        target_wwn = None
        for port in fc_ports:
            nsp = port['nsp']
            if target_nsp == nsp:
                target_wwn = port['portWWN']
                break

        if not target_wwn:
            LOG.warning("Did not get wwn for target nsp: "
                        "%(nsp)s", {'nsp': target_nsp})

        return target_wwn

    def merge_dicts(self, dict_1, dict_2):
        keys = set(dict_1).union(dict_2)
        no = []
        return {k: (dict_1.get(k, no) + dict_2.get(k, no)) for k in keys}



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from cinder import context
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.drivers.dell_emc import powerflex


class TestInitializeConnection(powerflex.TestPowerFlexDriver):
    def setUp(self):
        """Setup a test case environment."""

        super(TestInitializeConnection, self).setUp()
        self.connector = {}
        self.ctx = (
            context.RequestContext('fake', 'fake', True, auth_token=True))
        self.volume = fake_volume.fake_volume_obj(
            self.ctx, **{'provider_id': fake.PROVIDER_ID})

    def test_only_qos(self):
        qos = {'maxIOPS': 1000, 'maxBWS': 2048}
        extraspecs = {}
        connection_properties = (
            self._initialize_connection(qos, extraspecs)['data'])
        self.assertEqual(1000, int(connection_properties['iopsLimit']))
        self.assertEqual(2048, int(connection_properties['bandwidthLimit']))

    def test_no_qos(self):
        qos = {}
        extraspecs = {}
        connection_properties = (
            self._initialize_connection(qos, extraspecs)['data'])
        self.assertIsNone(connection_properties['iopsLimit'])
        self.assertIsNone(connection_properties['bandwidthLimit'])

    def test_qos_scaling_and_max(self):
        qos = {'maxIOPS': 100, 'maxBWS': 2048, 'maxIOPSperGB': 10,
               'maxBWSperGB': 128}
        extraspecs = {}
        self.volume.size = 8
        connection_properties = (
            self._initialize_connection(qos, extraspecs)['data'])
        self.assertEqual(80, int(connection_properties['iopsLimit']))
        self.assertEqual(1024, int(connection_properties['bandwidthLimit']))

        self.volume.size = 24
        connection_properties = (
            self._initialize_connection(qos, extraspecs)['data'])
        self.assertEqual(100, int(connection_properties['iopsLimit']))
        self.assertEqual(2048, int(connection_properties['bandwidthLimit']))

    def test_qos_scaling_no_max(self):
        qos = {'maxIOPSperGB': 10, 'maxBWSperGB': 128}
        extraspecs = {}
        self.volume.size = 8
        connection_properties = (
            self._initialize_connection(qos, extraspecs)['data'])
        self.assertEqual(80, int(connection_properties['iopsLimit']))
        self.assertEqual(1024, int(connection_properties['bandwidthLimit']))

    def test_qos_round_up(self):
        qos = {'maxBWS': 2000, 'maxBWSperGB': 100}
        extraspecs = {}
        self.volume.size = 8
        connection_properties = (
            self._initialize_connection(qos, extraspecs)['data'])
        self.assertEqual(1024, int(connection_properties['bandwidthLimit']))

        self.volume.size = 24
        connection_properties = (
            self._initialize_connection(qos, extraspecs)['data'])
        self.assertEqual(2048, int(connection_properties['bandwidthLimit']))

    def test_vol_id(self):
        extraspecs = qos = {}
        connection_properties = (
            self._initialize_connection(qos, extraspecs)['data'])
        self.assertEqual(fake.PROVIDER_ID,
                         connection_properties['scaleIO_volume_id'])

    def _initialize_connection(self, qos, extraspecs):
        self.driver._get_volumetype_qos = mock.MagicMock()
        self.driver._get_volumetype_qos.return_value = qos
        self.driver._get_volumetype_extraspecs = mock.MagicMock()
        self.driver._get_volumetype_extraspecs.return_value = extraspecs
        return self.driver.initialize_connection(self.volume, self.connector)
